"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const fs_extra_1 = __importDefault(require("fs-extra"));
const vueResolver_1 = require("./vueResolver");
const resolve_from_1 = __importDefault(require("resolve-from"));
const chalk_1 = __importDefault(require("chalk"));
const resolver_1 = require("./resolver");
const buildPluginResolve_1 = require("./buildPluginResolve");
const buildPluginHtml_1 = require("./buildPluginHtml");
const buildPluginCss_1 = require("./buildPluginCss");
const buildPluginAsset_1 = require("./buildPluginAsset");
const utils_1 = require("./utils");
async function build(options = {}) {
    process.env.NODE_ENV = 'production';
    const start = Date.now();
    const { root = process.cwd(), cdn = !vueResolver_1.resolveVue(root).hasLocalVue, outDir = path_1.default.resolve(root, 'dist'), assetsDir = 'assets', resolvers = [], srcRoots = [], rollupInputOptions = {}, rollupOutputOptions = {}, rollupPluginVueOptions = {}, emitAssets = true, write = true, minify = true, silent = false } = options;
    // lazy require rollup so that we don't load it when only using the dev server
    // importing it just for the types
    const rollup = require('rollup').rollup;
    const indexPath = path_1.default.resolve(root, 'index.html');
    const cssFileName = 'style.css';
    const resolvedAssetsPath = path_1.default.join(outDir, assetsDir);
    let indexContent = null;
    try {
        indexContent = await fs_extra_1.default.readFile(indexPath, 'utf-8');
    }
    catch (e) {
        // no index
    }
    const resolver = resolver_1.createResolver(root, resolvers);
    srcRoots.push(root);
    const bundle = await rollup({
        input: path_1.default.resolve(root, 'index.html'),
        ...rollupInputOptions,
        plugins: [
            // user plugins
            ...(rollupInputOptions.plugins || []),
            // vite:resolve
            buildPluginResolve_1.createBuildResolvePlugin(root, cdn, srcRoots, resolver),
            // vite:html
            ...(indexContent ? [buildPluginHtml_1.createBuildHtmlPlugin(indexPath, indexContent)] : []),
            // vue
            require('rollup-plugin-vue')({
                transformAssetUrls: {
                    includeAbsolute: true
                },
                // TODO: for now we directly handle pre-processors in rollup-plugin-vue
                // so that we don't need to install dedicated rollup plugins.
                // In the future we probably want to still use rollup plugins so that
                // preprocessors are also supported by importing from js files.
                preprocessStyles: true,
                preprocessCustomRequire: (id) => require(resolve_from_1.default(root, id)),
                // TODO proxy cssModules config
                ...rollupPluginVueOptions
            }),
            require('@rollup/plugin-json')(),
            require('@rollup/plugin-node-resolve')({
                rootDir: root
            }),
            require('@rollup/plugin-replace')({
                'process.env.NODE_ENV': '"production"',
                __DEV__: 'false'
            }),
            // vite:css
            buildPluginCss_1.createBuildCssPlugin(root, assetsDir, cssFileName, minify),
            // vite:asset
            buildPluginAsset_1.createBuildAssetPlugin(assetsDir),
            // minify with terser
            // modules: true and toplevel: true are implied with format: 'es'
            ...(minify ? [require('rollup-plugin-terser').terser()] : [])
        ],
        onwarn(warning, warn) {
            if (warning.code !== 'CIRCULAR_DEPENDENCY') {
                warn(warning);
            }
        }
    });
    const { output } = await bundle.generate({
        format: 'es',
        ...rollupOutputOptions
    });
    let generatedIndex = indexContent && indexContent.replace(buildPluginHtml_1.scriptRE, '').trim();
    const injectCSS = (html, filename) => {
        const tag = `<link rel="stylesheet" href="/${path_1.default.posix.join(assetsDir, filename)}">`;
        if (/<\/head>/.test(html)) {
            return html.replace(/<\/head>/, `${tag}\n</head>`);
        }
        else {
            return tag + '\n' + html;
        }
    };
    const injectScript = (html, filename) => {
        filename = utils_1.isExternalUrl(filename)
            ? filename
            : `/${path_1.default.posix.join(assetsDir, filename)}`;
        const tag = `<script type="module" src="${filename}"></script>`;
        if (/<\/body>/.test(html)) {
            return html.replace(/<\/body>/, `${tag}\n</body>`);
        }
        else {
            return html + '\n' + tag;
        }
    };
    // TODO handle public path for injections?
    // this would also affect paths in templates and css.
    if (generatedIndex) {
        // inject css link
        generatedIndex = injectCSS(generatedIndex, cssFileName);
        if (cdn) {
            // if not inlining vue, inject cdn link so it can start the fetch early
            generatedIndex = injectScript(generatedIndex, vueResolver_1.resolveVue(root).cdnLink);
        }
    }
    if (write) {
        await fs_extra_1.default.remove(outDir);
        await fs_extra_1.default.ensureDir(outDir);
    }
    // inject / write bundle
    for (const chunk of output) {
        if (chunk.type === 'chunk') {
            if (chunk.isEntry && generatedIndex) {
                // inject chunk to html
                generatedIndex = injectScript(generatedIndex, chunk.fileName);
            }
            // write chunk
            if (write) {
                const filepath = path_1.default.join(resolvedAssetsPath, chunk.fileName);
                !silent &&
                    console.log(`write ${chalk_1.default.cyan(path_1.default.relative(process.cwd(), filepath))}`);
                await fs_extra_1.default.ensureDir(path_1.default.dirname(filepath));
                await fs_extra_1.default.writeFile(filepath, chunk.code);
            }
        }
        else if (emitAssets && write) {
            // write asset
            const filepath = path_1.default.join(resolvedAssetsPath, chunk.fileName);
            !silent &&
                console.log(`write ${chalk_1.default.magenta(path_1.default.relative(process.cwd(), filepath))}`);
            await fs_extra_1.default.ensureDir(path_1.default.dirname(filepath));
            await fs_extra_1.default.writeFile(filepath, chunk.source);
        }
    }
    if (write) {
        // write html
        if (generatedIndex) {
            const indexOutPath = path_1.default.join(outDir, 'index.html');
            !silent &&
                console.log(`write ${chalk_1.default.green(path_1.default.relative(process.cwd(), indexOutPath))}`);
            await fs_extra_1.default.writeFile(indexOutPath, generatedIndex);
        }
    }
    !silent &&
        console.log(`Build completed in ${((Date.now() - start) / 1000).toFixed(2)}s.`);
    return {
        assets: output,
        html: generatedIndex || ''
    };
}
exports.build = build;
